package ap.proof;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, Conjunction, TermOrder, ConstantTerm, Quantifier}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.EquationConj
import ap.terfor.substitutions.{Substitution, IdentitySubst}
import ap.proof.goal.{Goal, FactsNormalisationTask, EliminateFactsTask, AddFactsTask}
import ap.util.{Debug, Logic}


/**
 * A prover that tries to construct a countermodel of a ground formula. This
 * prover works in depth-first mode, in contrast to the
 * <code>ExhaustiveProver</code>.
 */
object ModelSearchProver {

  private val AC = Debug.AC_PROVER
   
  private val simplifier = ConstraintSimplifier.NO_SIMPLIFIER
  
  // we need to store eliminated facts from goals, otherwise we could not
  // construct a complete model
  private val ptf = new SimpleProofTreeFactory(true, simplifier) {
    override def eliminatedConstant(subtree : ProofTree,
                                    c : ConstantTerm,
                                    witness : (Substitution, TermOrder) => Substitution,
                                    order : TermOrder) : ProofTree =
      new WitnessTree (subtree, c, witness, order)
  }

  /**
   * <code>inputFor</code> is the formula to be disproven. The result of the
   * method is a countermodel of <code>inputFor</code>, or <code>FALSE</code>
   * if it was not possible to find one (this implies that <code>inputFor</code>
   * is valid).
   */
  def apply(inputFor : Formula, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ModelSearchProver.AC, inputFor.variables.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val goal = Goal(inputFor, Set() ++ order.orderedConstants, order)
    try {
      expandProof(goal)
      Conjunction.FALSE // no model has been found
    } catch {
      case FoundModel(facts, constantValues, order) => {
        val valuesConj =
          EquationConj(for (c <- order.orderedConstants.elements)
                       yield LinearCombination(Array((IdealInt.ONE, c),
                                                     (IdealInt.MINUS_ONE, constantValues(c))),
                                               order),
                       order)
        Conjunction.conj(Array(valuesConj, facts), order)
      }
    }
  }

  private def expandProof(tree : ProofTree) : unit =
    if (tree.closingConstraint.isFalse) tree match { 

      case goal : Goal => {
        if (goal.stepPossible)
          expandProof(goal.step(ptf))
        else
          throw FoundModel(goal.facts,
                           new IdentitySubst (goal.order),
                           goal.order)
      }

      case WitnessTree(subtree, eliminatedConstant, witness, order) =>
        try {
          expandProof(subtree)
          ////////////////////////////////////////
          // some arbitrary operations to avoid tail recursion ...
          // this seems to be a compiler bug
          // TODO: write a bug report
          var x = 0; x = x
        } catch {
          case FoundModel(facts, constantValues, order) => {
            throw FoundModel(facts,
                             witness(constantValues, order),
                             order)
          }
        }
      
      case tree : ProofTreeOneChild => {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(ModelSearchProver.AC, tree match {
                          case _ : WeakenTree => false
                          case tree : QuantifiedTree => tree.quan == Quantifier.ALL
                          case _ => true
                        })
        ////////////////////////////////////////////////////////////////////////
        expandProof(tree.subtree)
      }
     
      case tree : AndTree => {
        expandProof(tree.left)
        expandProof(tree.right)
      }
      
    }
  
  private case class FoundModel(facts : Conjunction,
                                constantValues : Substitution,
                                order : TermOrder) extends Exception

}

 
private case class WitnessTree(val subtree : ProofTree,
                               val eliminatedConstant : ConstantTerm,
                               val witness : (Substitution, TermOrder) => Substitution,
                               val order : TermOrder) extends ProofTreeOneChild {
  
  def update(newSubtree : ProofTree) : ProofTree =
    new WitnessTree(subtree, eliminatedConstant, witness, order)

  lazy val closingConstraint : Formula = subtree.closingConstraint

  lazy val stepMeaningful : Boolean = subtree.stepMeaningful

  protected val subtreeOrder = order
  
}