package ap.proof.goal;

import scala.collection.mutable.ArrayBuffer

import ap.basetypes.IdealInt
import ap.terfor.{TerFor, Term, Formula, ConstantTerm, Conjunction, TermOrder}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.arithconj.{ArithConj, ModelFinder}
import ap.terfor.inequalities.InEqConj
import ap.terfor.substitutions.Substitution
import ap.util.{Logic, Debug, Seqs, FilterIt}

/**
 * Task for removing facts that are no longer needed (like equations that have
 * been applied to all other formulas), or that can be discharged directly by
 * moving them into the contraint.
 */
case object EliminateFactsTask extends EagerTask {

  val AC = Debug.AC_ELIM_FACTS_TASK

  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val eliminator = new Eliminator(goal.facts, goal, ptf)

    ////////////////////////////////////////////////////////////////////////////
    
    if (eliminator.facts == goal.facts)
      // nothing has changed and the application of this task was unnecessary
      ptf.updateGoal(goal)
    else
      eliminator.postProcessor(ptf.updateGoal(eliminator.facts, goal))
  }

}


private class Eliminator(oriFacts : Conjunction,
                         goal : Goal, ptf : ProofTreeFactory) {
  
  var facts = oriFacts
  private val order = goal.order
  private val eliminatedConstants = goal.eliminatedConstants
  
  var postProcessor : ProofTree => ProofTree = ((p) => p)

  private def contains(set : Set[ConstantTerm], c : Term) = c match {
    case c : ConstantTerm => set contains c
    case _ => false
  }

  private def occursInPositiveEqs(c : Term) =
    contains(facts.arithConj.positiveEqs.constants, c)

  private def occursInNegativeEqs(c : Term) =
    contains(facts.arithConj.negativeEqs.constants, c)

  private def occursInInEqs(c : Term) =
    contains(facts.arithConj.inEqs.constants, c)

  private def occursInTasks(c : Term) =
    contains(goal.tasks.taskInfos.constants, c) || contains(facts.predConj.constants, c)

  private def containsEliminatedConstants(f : TerFor) =
    !Seqs.disjoint(f.constants, eliminatedConstants)
  
  private def addConstraint(f : Formula) =
    postProcessor = postProcessor compose
      ((pt:ProofTree) => ptf.weaken(pt, f, order))      
  
  private def addModelInformation(eliminatedConstant : ConstantTerm,
                                  witness : (Substitution, TermOrder) => Substitution) =
    postProcessor = postProcessor compose
      ((pt:ProofTree) => ptf.eliminatedConstant(pt, eliminatedConstant,
                                                witness, order))      

  ////////////////////////////////////////////////////////////////////////////
  // Positive equations
    
  private def eliminablePositiveEqs(c : ConstantTerm) : Boolean = {
    var occurred : Boolean = false
    for (lc <- facts.arithConj.positiveEqs)
      if (lc.constants contains c) {
        // the constant must occur in at most one equation
        if (occurred) return false
        occurred = true

        // and the coefficient must be at most one
        if ((lc get c).abs > IdealInt.ONE) return false
        
        // we do not remove the equation if c is not an eliminated constant, but
        // the equation contains other eliminated constants;
        // there are chances that we can remove the equation completely later
        if (!(eliminatedConstants contains c) &&
            containsEliminatedConstants(lc)) return false        
      }
    
    true
  }
    
  private def elimPositiveEqs(c : ConstantTerm) : unit = {
    val remainingEqs = new ArrayBuffer[LinearCombination]
   
    for (lc <- facts.arithConj.positiveEqs)
      if (lc.constants contains c) {        
        if (eliminatedConstants contains c) {
          // then we can just ignore the equation; we describe how to compute
          // a witness for the eliminated constant c
          val modelFinder =
            new ModelFinder (ArithConj.conj(EquationConj(lc, order), order), c)
          addModelInformation(c, modelFinder)
        } else {
          // the equation can directly be moved to the constraint
          addConstraint(NegEquationConj(lc, order))
        }
      } else {
        remainingEqs += lc
      }

    facts = facts.updatePositiveEqs(EquationConj(remainingEqs, order), order)
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Eliminated constants that occur in negative equations. The result are the
  // eliminated equations
    
  private def elimNegativeEqsU(c : ConstantTerm) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(EliminateFactsTask.AC, eliminatedConstants contains c)
    ////////////////////////////////////////////////////////////////////////////
    val eqs = facts.arithConj.negativeEqs
    
    val remainingEqs = new ArrayBuffer[LinearCombination]
    val eliminatedEqs = new ArrayBuffer[LinearCombination]
    for (lc <- eqs)
      if (lc.constants contains c) {        
        // then we can eliminate the equation
	eliminatedEqs += lc
      } else {
        remainingEqs += lc
      }
    
    facts = facts.updateNegativeEqs(eqs.updateEqs(remainingEqs, order), order)

    // we give back the eliminated equations
    NegEquationConj(eliminatedEqs, order)
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Eliminated constants that occur in inequalities

  private def onesidedInEqsU(c : ConstantTerm) : Boolean = {
    // the coefficient of the constant must have the same sign in all inequalities
    
    var signum : Int = 0
    for (lc <- facts.arithConj.inEqs) {
      val newSignum = (lc get c).signum
      if (newSignum != 0) {
        if (signum * newSignum == -1) return false
        signum = newSignum
      }
    }
    
    true
  }

  private def elimOnesidedInEqsU(c : ConstantTerm) : InEqConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(EliminateFactsTask.AC, eliminatedConstants contains c)
    ////////////////////////////////////////////////////////////////////////////
    val inEqs = facts.arithConj.inEqs
    
    val remainingInEqs = new ArrayBuffer[LinearCombination]
    val eliminatedInEqs = new ArrayBuffer[LinearCombination]
    for (lc <- inEqs)
      if (lc.constants contains c) {
        // then we can eliminate the inequality
        eliminatedInEqs += lc
      } else {
        remainingInEqs += lc
      }

    facts = facts.updateInEqs(inEqs.updateGeqZero(remainingInEqs, order), order)

    // we give back the eliminated inequalities
    InEqConj(eliminatedInEqs, order)
  }
    
  //////////////////////////////////////////////////////////////////////////////
  // Non-eliminated constants that occur in negative equations

  private def eliminableNegativeEqs(c : ConstantTerm) : Boolean =
    // we only move equations to the constraints if no eliminated
    // constants occur in any of them
    Logic.forall(for (lc <- facts.arithConj.negativeEqs.elements)
                 yield (!(lc.constants contains c) || !containsEliminatedConstants(lc)))

  private def elimNegativeEqs(c : ConstantTerm) : unit = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(EliminateFactsTask.AC, !(eliminatedConstants contains c))
    ////////////////////////////////////////////////////////////////////////////

    val remainingEqs = new ArrayBuffer[LinearCombination]
    val constraintEqs = new ArrayBuffer[LinearCombination]
      
    for (lc <- facts.arithConj.negativeEqs)
      if (lc.constants contains c)
        constraintEqs += lc
      else
        remainingEqs += lc
      
    addConstraint(Conjunction.disj(for (lc <- constraintEqs.elements)
                                   yield EquationConj(lc, order),
                  order))
    facts = facts.updateNegativeEqs(NegEquationConj(remainingEqs, order), order)
  }  

  //////////////////////////////////////////////////////////////////////////////
  // Exact Fourier-Motzkin elimination (exact shadow)
  
  private def eliminableExactShadow(c : ConstantTerm) : Boolean = {
    // we check that either all lower or all upper bounds have the leading
    // coefficient one; otherwise, Fourier-Motzkin is not precise
    var lowerNonUnit : Boolean = false
    var upperNonUnit : Boolean = false
    
    for (lc <- facts.arithConj.inEqs) if (!lc.isEmpty) {
      if (c == lc.leadingTerm) {
        if (lc.leadingCoeff > IdealInt.ONE) lowerNonUnit = true
        if (lc.leadingCoeff < IdealInt.MINUS_ONE) upperNonUnit = true
      } else {
        // if c occurs as a non-leading term, elimination is not possible
        if (lc.constants contains c) return false
      }
    }
    
    !lowerNonUnit || !upperNonUnit
  }

  private def elimExactShadow(c : ConstantTerm) : InEqConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(EliminateFactsTask.AC, eliminatedConstants contains c)
    ////////////////////////////////////////////////////////////////////////////

    val inEqs = facts.arithConj.inEqs    
    facts = facts.updateInEqs(inEqs.exactShadow(c, order), order)
    
    // we give back the eliminated inequalities
    inEqs.updateGeqZero(inEqs filter ((lc) => c == lc.leadingTerm), order)
  }

  //////////////////////////////////////////////////////////////////////////////
  // The main loop
    
  var oldFacts = facts
  do {
    oldFacts = facts

    for (c <- facts.constants) if (!occursInTasks(c)) { // otherwise we cannot do anything
      (occursInPositiveEqs(c), occursInNegativeEqs(c), occursInInEqs(c)) match {

      case (false, false, false) => // nothing
      
      case (true, false, false)
        if (eliminablePositiveEqs(c))
          => elimPositiveEqs(c)
   
      case (false, true, false)
        if (!(eliminatedConstants contains c) && eliminableNegativeEqs(c))
          => elimNegativeEqs(c)

      case (false, false, true)
        if ((eliminatedConstants contains c) && eliminableExactShadow(c))
          => {
               val eliminatedFor = ArithConj.conj(elimExactShadow(c), order)
               addModelInformation(c, new ModelFinder (eliminatedFor, c))
             }

      case (false, _, _)
        if ((eliminatedConstants contains c) && onesidedInEqsU(c))
          => {
               val eliminatedFor = ArithConj.conj(Array(elimNegativeEqsU(c),
                                                        elimOnesidedInEqsU(c)),
                                                  order)
               addModelInformation(c, new ModelFinder (eliminatedFor, c))
             }
    
      case _ => // nothing
        
      }
    }
    
  } while (oldFacts != facts)

}