package ap.proof.goal;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.{Conjunction, Formula, ConstantTerm, VariableTerm,
                  TermOrder, TerFor, Term, Quantifier}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.substitutions.ConstantSubst
import ap.util.{Debug, FilterIt, Logic}

object Goal {
  
  private val AC = Debug.AC_GOAL

  def apply(facts : Conjunction, tasks : TaskManager, age : Int,
            eliminatedConstants : Set[ConstantTerm], order : TermOrder) : Goal = {
    if (facts.isFalse) {
      TRUE
    } else {
      new Goal(facts, tasks, age, eliminatedConstants, order)
    }
  }
  
  def apply(f : Formula, order : TermOrder) : Goal =
    apply(Conjunction.TRUE,
          TaskManager.EMPTY ++ FormulaTask(Conjunction.conj(f, order), 0),
          0,
          Set.empty,
          order)
  
  def apply(f : Formula,
            eliminatedConstants : Set[ConstantTerm],
            order : TermOrder) : Goal =
    apply(Conjunction.TRUE,
          TaskManager.EMPTY ++ FormulaTask(Conjunction.conj(f, order), 0),
          0,
          eliminatedConstants,
          order)
  
  val TRUE : Goal = new Goal (Conjunction.FALSE, TaskManager.EMPTY, 0,
                              Set.empty, TermOrder.EMPTY)
  
}

class Goal private (val facts : Conjunction,
                    val tasks : TaskManager,
                    val age : Int,
                    val eliminatedConstants : Set[ConstantTerm],
                    val order : TermOrder) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Goal.AC,
                   // facts are really just literals (equations, inequalities,
                   // predicate literals)
                   facts.quans.isEmpty && facts.negatedConjs.isTrue &&
                   // there should not be any free variables (use constants
                   // instead)
                   facts.variables.isEmpty &&
                   // if a contradiction is detected, we can forget about
                   // everything
                   (!facts.isFalse || tasks.isEmpty) &&
                   // we must choose an ordering such that the constants to be
                   // eliminated are maximal
                   (facts isSortedBy order) &&
                   (order constantsAreMaximal eliminatedConstants))
  //////////////////////////////////////////////////////////////////////////////

  private def elimConstants(tfs : Iterator[LinearCombination])
                                                : Iterator[LinearCombination] =
    FilterIt(tfs, (tf:LinearCombination) => {
      val res = tf.isEmpty || !(this eliminates tf.leadingTerm)
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(Goal.AC,
                       res == (tf.constants intersect eliminatedConstants).isEmpty)
      //////////////////////////////////////////////////////////////////////////
      res
    })
    
  lazy val closingConstraint : Formula = {
    val ac = facts.arithConj
    
    Conjunction.negate(
      ArithConj(EquationConj(elimConstants(ac.positiveEqs.elements), order),
                NegEquationConj(elimConstants(ac.negativeEqs.elements), order),
                InEqConj(elimConstants(ac.inEqs.elements ++
                                       ac.inEqs.geqZeroInfs.elements),
                         order),
                order),
      order)
  }
  
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return whether <code>t</code> is a constant that is eliminated in this goal
   */
  def eliminates(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  //////////////////////////////////////////////////////////////////////////////

  val stepPossible : Boolean = !tasks.isEmpty

  val stepMeaningful : Boolean = stepPossible
  
  def step(ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Goal.AC, this.stepPossible)
    ////////////////////////////////////////////////////////////////////////////
    
    // apply the first task to the goal and then stop immediately
    tasks.max(this, ptf)
  }

  //////////////////////////////////////////////////////////////////////////////
    
  override def toString : String = {
    "Facts: " + facts + "\n" +
    "Tasks: " + tasks + "\n" +
    "Eliminated constants: " + eliminatedConstants + "\n" +
    "Constraint: " + closingConstraint
  }
}
