package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Term, TermOrder, Conjunction, ConstantTerm}
import ap.terfor.inequalities.InEqConj
import ap.terfor.linearcombination.LinearCombination
import ap.util.{Logic, Debug, Seqs}


/**
 * A task for splitting equations in the succedent into two inequalities. This
 * can be necessary if the equation contains the same symbols as other
 * inequalities
 */
case object SplitEqTask extends EagerTask {

  private val AC = Debug.AC_FACTS_TASK

  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val ac = goal.facts.arithConj
    val order = goal.order
    
    for (eq <- ac.negativeEqs) {
      if (!eq.isEmpty && (goal eliminates eq.leadingTerm)) {
        val lt = eq.leadingTerm
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(SplitEqTask.AC,
                        lt.isInstanceOf[ConstantTerm] &&
                        // this task should only be applied very late, when all
                        // other tasks have already been done
                        !(goal.tasks.taskInfos.constants contains lt.asInstanceOf[ConstantTerm]))
        ////////////////////////////////////////////////////////////////////////
        if (containsLT(lt, ac.inEqs.geqZero, order) ||
            containsLT(lt, ac.inEqs.geqZeroInfs, order)) {
          // The equation has to be split. We have to split the goal directly,
          // because when just adding a formula eq >= 0 & eq <= 0 this will
          // directly be simplified to an equation
    
          // TODO: directly remove the negated equation
          
          val lowerBound = Conjunction.conj(InEqConj(eq, order), order)
          val upperBound = Conjunction.conj(InEqConj(-eq, order), order)
          val goal1 = ptf.updateGoal(FormulaTask(lowerBound, goal.age), goal)
          val goal2 = ptf.updateGoal(FormulaTask(upperBound, goal.age), goal)
          
          return ptf.and(Array(goal1, goal2), order)
        }
      } else {
        ////////////////////////////////////////////////////////////////////////
        Debug.assertInt(SplitEqTask.AC,
                        Seqs.disjoint(goal.eliminatedConstants, eq.constants))
        ////////////////////////////////////////////////////////////////////////
      }
    }
    
    // nothing to be done
    ptf updateGoal goal
  }

  /**
   * Determine whether a sequence of <code>LinearCombination</code>
   * (sorted in descending order) contains an element with a certain leading
   * term
   */
  private def containsLT(lt : Term,
                         seq : RandomAccessSeq[LinearCombination],
                         order : TermOrder) : Boolean = {
    implicit def orderLC(thisLC : LinearCombination) =
      new Ordered[LinearCombination] {
        def compare(thatLC : LinearCombination) : Int =
          order.compare(thatLC, thisLC)
      }
  
    Seqs.binSearch(seq, 0, seq.size, LinearCombination(lt, order)) match {
    case Seqs.Found(_) => true
    case Seqs.NotFound(i) =>
      (i < seq.size) && !seq(i).isEmpty && (seq(i).leadingTerm == lt) ||
      (i > 0) && !seq(i-1).isEmpty && (seq(i-1).leadingTerm == lt)
    }
  }

}
