package ap.terfor;

import ap.util.{CountIt, Logic, Seqs, Debug, FilterIt}
import ap.basetypes.IdealInt
import linearcombination.LinearCombination
import arithconj.ArithConj
import preds.{Predicate, Atom}

/**
 * Class for representing total orderings on constants (and variables), and
 * their extension to arbitrary terms. For the time being, we do not consider
 * arbitrary (non-nullary) functions apart from the pre-defined arithmetic
 * operations.
 *
 * <code>constantSeq</code> is the list of constants that are totally ordered by
 * this <code>TermOrder</code>, starting with the smallest constant. In
 * addition, variable terms are by default ordered as bigger as all constants.
 */
object TermOrder {

  private val AC = Debug.AC_TERM_ORDER

  /** The term order that cannot sort anything apart from
   * <code>Constant.ONE</code> and variables */
  val EMPTY = new TermOrder(List(), List())
  
}
   
class TermOrder private (private val constantSeq : Seq[ConstantTerm],
                         private val predicateSeq : Seq[Predicate]) {

  private val constantWeight : scala.collection.Map[ConstantTerm, NonCoeffWeight] = {
    val res = new scala.collection.mutable.HashMap[ConstantTerm, NonCoeffWeight]
    res ++= (constantSeq.elements zip
             (for (i <- new CountIt) yield ConstantWeight(i)))
    res
  }
  
  private val predicateWeight : scala.collection.Map[Predicate, Int] = {
    val res = new scala.collection.mutable.HashMap[Predicate, Int]
    res ++= predicateSeq.elements.zipWithIndex
    res
  }
  
  //////////////////////////////////////////////////////////////////////////////
  private def noDuplicates[A](seq : Seq[A]) : Boolean =
    Logic.forall(0, seq.size, (i:Int) =>
    Logic.forall(i+1, seq.size, (j:Int) =>
                 constantSeq(i) != constantSeq(j)))
  Debug.assertCtor(TermOrder.AC,
                   noDuplicates(constantSeq) && noDuplicates(predicateSeq))
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Test whether <code>x</code> is sorting by this <code>TermOrder</code>, or
   * return <code>true</code> if <code>x</code> is not a sorted entity
   */
  def isSortingOf[A](x : A) : Boolean = x match {
    case x : Sorted[A] => x isSortedBy this
    case _ => true
  }

  /**
   * If <code>x</code> is a sorted entity, sort it by this
   * <code>TermOrder</code>, or otherwise return the unchanged <code>x</code>
   */
  def sort[A](x : A) : A = x match {
    case x : Sorted[A] => x sortBy this
    case _ => x
  }

  //////////////////////////////////////////////////////////////////////////////

  def compare(t1 : Term, t2 : Term) : Int =
    Seqs.lexCompare(weightIt(t1), weightIt(t2))

  def compare(c1 : Seq[LinearCombination], c2 : Seq[LinearCombination]) : Int = {
    implicit def orderLCs(thisLC : LinearCombination) =
      new Ordered[LinearCombination] {
        def compare(thatLC : LinearCombination) : Int =
          TermOrder.this.compare(thisLC, thatLC)
      }
      
    Seqs.lexCompare(c1.elements, c2.elements)
  }

  def compare(c1 : ArithConj, c2 : ArithConj) : Int =
    Seqs.lexCombineInts(compare(c1.positiveEqs, c2.positiveEqs),
                        compare(c1.negativeEqs, c2.negativeEqs),
                        compare(c1.inEqs, c2.inEqs))
    
  //////////////////////////////////////////////////////////////////////////////

  def compare(a1 : Atom, a2 : Atom) : Int = {
    implicit def orderLCs(thisLC : LinearCombination) =
      new Ordered[LinearCombination] {
        def compare(thatLC : LinearCombination) : Int =
          TermOrder.this.compare(thisLC, thatLC)
      }

    Seqs.lexCombineInts(predicateWeight(a1.pred) compare predicateWeight(a2.pred),
                        Seqs.lexCompare(a1.elements, a2.elements))
  }
    
  //////////////////////////////////////////////////////////////////////////////

  /**
   * The comparison of terms is reduced to the lexicographic comparison of
   * <code>Weight</code> objects
   */
  private def weightIt(t : Term) : Iterator[Weight] = t match {
    case t : LinearCombination => (for ((coeff, ter) <- t.elements)
                                   yield CoeffWeight(coeff, weightOfAtomicTerm(ter)))
    case _ => Iterator.single(weightOfAtomicTerm(t))
  }

  private def weightOfAtomicTerm(t : Term) : NonCoeffWeight = t match {
    case c : ConstantTerm => constantWeight(c)
    case VariableTerm(i) => VariableWeight(-i)
    case OneTerm => OneWeight
  }
  
  /**
   * Return <code>true</code> iff the term order <code>that</code> is an
   * extension of the order <code>this</code>, i.e., when restricted to the
   * constants that are ordered by <code>this</code> it describes the same
   * order.
   */
  def isSubOrderOf(that : TermOrder) : Boolean = 
    isSubOrderOf(constantSeq.elements, that) &&
    subSeq(this.predicateSeq.elements, that.predicateSeq.elements)

  /**
   * Return <code>true</code> iff the term order <code>that</code> is an
   * extension of the order <code>this</code> considering only the constants
   * <code>consideredConstants</code>. I.e., <code>this</code> restricted to
   * the constants in <code>consideredConstants</code> is a suborder of
   * <code>that</code>.
   */
  def isSubOrderOf(that : TermOrder,
                   consideredConstants : scala.collection.Set[ConstantTerm]) : Boolean = 
    isSubOrderOf(FilterIt(constantSeq.elements,
                          (c:ConstantTerm) => consideredConstants contains c),
                 that) &&
    subSeq(this.predicateSeq.elements, that.predicateSeq.elements)

  private def isSubOrderOf(thisSeq : Iterator[ConstantTerm], that : TermOrder)
                                                           : Boolean =
    subSeq(thisSeq, that.constantSeq.elements)
   
  private def subSeq[A](a : Iterator[A], b : Iterator[A]) : Boolean = {
    for (const <- a) {
      do {
        if (!b.hasNext) return false
      } while (const != b.next)
    }
    true    
  }
   
  /**
   * Extend this ordering by inserting a further constant <code>newConst</code>.
   * This constant is inserted so that it gets as big as possible, but such that
   * it is smaller than all constants of the set <code>biggerConstants</code>
   */
  def extend(newConst : ConstantTerm,
             biggerConstants : scala.collection.Set[ConstantTerm]) : TermOrder = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(TermOrder.AC,
                    !(constantWeight contains newConst) &&
                    !(biggerConstants contains newConst))
    ////////////////////////////////////////////////////////////////////////////

    val newConstantSeq = new Array[ConstantTerm](constantSeq.size + 1)
    var pos : Int = 0
    var inserted : Boolean = false
    
    for (const <- constantSeq) {
      if (!inserted && (biggerConstants contains const)) {
        newConstantSeq(pos) = newConst
        pos = pos + 1
        inserted = true
      }
      newConstantSeq(pos) = const
      pos = pos + 1
    }
    
    if (!inserted) newConstantSeq(pos) = newConst
    
    val res = new TermOrder (newConstantSeq, predicateSeq)

    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(TermOrder.AC,
                     res.constantSeq.size == constantSeq.size + 1 &&
                     (this isSubOrderOf res) &&
                     Logic.exists(0, res.constantSeq.size,
                       (i:Int) => res.constantSeq(i) == newConst &&
                          Logic.forall(0, i,
                                       (j:Int) => !(biggerConstants contains
                                                    res.constantSeq(j)))))
    ////////////////////////////////////////////////////////////////////////////

    res
  }

  /**
   * Extend this ordering by inserting a further predicate <code>newPred</code>.
   * This predicate is inserted so that it gets as big as possible
   */
  def extend(newPred : Predicate) : TermOrder =
    new TermOrder(constantSeq, predicateSeq ++ List(newPred))

  /**
   * Determine whether this term order does not consider any constants as bigger
   * than the given constants 
   */
  def constantsAreMaximal(consts: Set[ConstantTerm]) : Boolean = {

    ////////////////////////////////////////////////////////////////////////////
    def post(b : Boolean) = {
      Debug.assertPost(TermOrder.AC,
                       b != (Logic.exists(0, constantSeq.size, (i) =>
                             Logic.exists(i+1, constantSeq.size, (j) =>
                               (consts contains constantSeq(i)) &&
                               !(consts contains constantSeq(j))))))
      b
    }
    ////////////////////////////////////////////////////////////////////////////
    
    var elem : Boolean = false
    for (c <- constantSeq) {
      val nextElem = consts contains c
      if (elem && !nextElem) return post(false)
      elem = nextElem
    }
    post(true)
  }
   
  /**
   * Return the set of all constants that are sorted by this
   * <code>TermOrder</code>
   */
  lazy val orderedConstants : scala.collection.Set[ConstantTerm] = {
    val res = new scala.collection.mutable.HashSet[ConstantTerm]
    res ++= constantSeq
    res
  }
   
  override def equals(that : Any) : Boolean = that match {
   case that : TermOrder => (this.constantSeq sameElements that.constantSeq) &&
                            (this.predicateSeq sameElements that.predicateSeq)
   case _ => false
  }
 
  private lazy val hashCodeVal =
    Seqs.computeHashCode(this.constantSeq, 1789817, 3) +
    Seqs.computeHashCode(this.predicateSeq, 178283, 5)

  override def hashCode = hashCodeVal

}

 
/**
 * Weight objects that are used to abstract from the concrete terms. There are
 * three types of weights: for variables, and for constants, and for the
 * literal 1. Variables are heavier that constants are heavier than literals.
 */
private abstract class Weight extends Ordered[Weight] {
  def compare(that : Weight) = (this.toCoeffWeight compare that.toCoeffWeight)
  
  protected def toCoeffWeight : CoeffWeight
}   
   
private abstract class NonCoeffWeight extends Weight {
  protected def patternType : Int
  protected def value : Int
  
  def compare(that : NonCoeffWeight) = 
    Seqs.lexCombineInts(this.patternType compare that.patternType,
                        this.value compare that.value)

  protected def toCoeffWeight : CoeffWeight = CoeffWeight(IdealInt.ONE, this)
}

private case object OneWeight extends NonCoeffWeight {
  protected def patternType : Int = 0
  protected def value : Int = 0  
}

private case class ConstantWeight(value : Int) extends NonCoeffWeight {
  protected def patternType : Int = 1
}

private case class VariableWeight(value : Int) extends NonCoeffWeight {
  protected def patternType : Int = 2
}

private case class CoeffWeight(coefficient : IdealInt, baseWeight : NonCoeffWeight)
                                                               extends Weight {
  def compare(that : CoeffWeight) : Int =
    Seqs.lexCombineInts(this.baseWeight compare that.baseWeight,
                        this.coefficient compareAbs that.coefficient)

  protected def toCoeffWeight : CoeffWeight = this
}